#include <nds.h>
#include <nds/arm9/console.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <user_debugger.h>
#include <fat.h> 

#include "sys.h"

#include "quake_ipc.h"

//#define USE_DEBUGGER
#define USE_3D
//#define USE_MP3
//#define USE_QUAKE_CONSOLE

#ifdef USE_DEBUGGER
#define DEBUG_MSG(x) debug_msg(x)
#else
#define DEBUG_MSG(x)
#endif

extern void quake_main (int argc, char **argv);

extern "C" void debug_print_stub(char* string)
{
	printf(string);
}

volatile unsigned int decoder_stopped = 1;
volatile unsigned int in_file_stuff = 0;

void file_lock(void)
{
	int lock_count = 0;
	while (in_file_stuff == 1)
	{
		if (lock_count > 1000000)
		{
			printf("trying to lock file access\n");
			lock_count = 0;
		}
		lock_count++;
	}
	
	in_file_stuff = 1;
//	printf("file access locked\n");
}

void file_unlock(void)
{
//	printf("file access unlocked\n");
	in_file_stuff = 0;
}

void handle_ipc(void)
{
	if (in_file_stuff)
		return;
	
	if (quake_ipc_7to9->message == 0xffffffff)
	{
		switch (quake_ipc_7to9->message_type)
		{
			case kPrintMessage:
			{
				printf("%s", quake_ipc_7to9_buf);
				break;
			}
			case kPrintNumber:
			{
				printf("ARM7: %08x", *(unsigned int *)quake_ipc_7to9_buf);
				break;
			}
			case kMalloc:
			{
				printf("ARM7 malloc req %d bytes\n", *(unsigned int *)quake_ipc_7to9_buf);
				unsigned int addr = (unsigned int)malloc(*(unsigned int *)quake_ipc_7to9_buf);
				
				if (addr == 0)
					printf("ARM7 MALLOC FAILED!\n");
				
				addr |= 0x400000;
				
				quake_ipc_9to7->message_type = kMalloc;
				*(unsigned int *)quake_ipc_9to7_buf = addr;
				quake_ipc_9to7->message = 0xffffffff;
				break;
			}
			case kFree:
			{
				printf("ARM7 free req on %08x\n", *(unsigned int *)quake_ipc_7to9_buf);
				void *ptr = (void *)*(unsigned int *)quake_ipc_7to9_buf;
				free(ptr);
				break;
			}
			case kFOpen:
			{
				file_lock();
				
				printf("ARM7 fopen: %s\n", quake_ipc_7to9_buf);
				FILE *fp = fopen((char *)quake_ipc_7to9_buf, "rb");
				
				if (fp)
					printf("...ok\n");
				else
					printf("...failed\n");
				
				printf("9: fp is %08x\n", (unsigned int)fp);
				
				*(unsigned int *)quake_ipc_9to7_buf = (unsigned int)fp;
				
				file_unlock();
				
				break;
			}
			case kFClose:
			{
				file_lock();
				
				printf("ARM7 fclose: %08x\n", *(unsigned int *)quake_ipc_7to9_buf);
				unsigned int fp = *(unsigned int *)quake_ipc_7to9_buf;
				fclose((FILE *)fp);
				
				file_unlock();
				
				break;
			}
			case kFRead:
			{
				file_lock();
				
				unsigned int buf = ((unsigned int *)quake_ipc_7to9_buf)[0];
				unsigned int size = ((unsigned int *)quake_ipc_7to9_buf)[1];
				unsigned int num = ((unsigned int *)quake_ipc_7to9_buf)[2];
				unsigned int fp = ((unsigned int *)quake_ipc_7to9_buf)[3];
				
					printf("ARM7 fread(%08x %d %d %08x)\n",
						buf, size, num, fp);
				
				unsigned int result = fread((unsigned char *)buf, size, num, (FILE *)fp);
				*(unsigned int *)quake_ipc_9to7_buf = result;
				
				file_unlock();
				
				break;
			}
			case kStopMP3:
			{
				printf("decoder has stopped\n");
				decoder_stopped = 1;
				break;
			}
		}
		
		quake_ipc_7to9->message = 0;
	}
}

/*
===============================================================================

FILE IO

===============================================================================
*/

#define MAX_HANDLES             10
FILE    *sys_handles[MAX_HANDLES];
/*
FILE *push_out_handle;
void push_out_init(char *filename)
{
	push_out_handle = fopen(filename, "wb+");
}

void push_out_end(void)
{
	fclose(push_out_handle);
}

void push_out(void *buf, int length, int pos)
{
	if (fseek(push_out_handle, pos, SEEK_SET) != 0)
		Sys_Error("push out seek failed\n");
	
	int res = fwrite(buf, length, 1, push_out_handle);
	
	if (res != length)
		Sys_Error("didn\'t write enough on push out, %d\n", res);
}

void pull_in(void *buf, int length, int pos)
{
	if (fseek(push_out_handle, pos, SEEK_SET) != 0)
		Sys_Error("pull in seek failed\n");
	
	int res = fread(buf, length, 1, push_out_handle);
	
	if (res != length)
		Sys_Error("didn\'t read enough on pull in, %d\n", res);
}*/

int             findhandle (void)
{
	int             i;
	
	for (i=1 ; i<MAX_HANDLES ; i++)
		if (!sys_handles[i])
			return i;
	Sys_Error ("out of handles");
	return -1;
}


/*
================
filelength
================
*/
int filelength (FILE *f)
{
	file_lock();
	
	int             pos;
	int             end;

	pos = ftell (f);
	fseek (f, 0, SEEK_END);
	end = ftell (f);
	fseek (f, pos, SEEK_SET);

	file_unlock();
	
	return end;
}

int Sys_FileOpenRead (char *path, int *hndl)
{
	FILE    *f;
	int             i;
	
	i = findhandle ();

	file_lock();
	f = fopen(path, "r");
	file_unlock();
	
	if (!f)
	{
		*hndl = -1;
		file_unlock();
		return -1;
	}
	sys_handles[i] = f;
	*hndl = i;
	
	int l = filelength(f);
	
	return l;
}

int Sys_FileOpenWrite (char *path)
{
	file_lock();
	
	FILE    *f;
	int             i;
	
	i = findhandle ();

	f = fopen(path, "w");
	if (!f)
		Sys_Error ("Error opening %s", path);
	sys_handles[i] = f;
	
	file_unlock();
	
	return i;
}

void Sys_FileClose (int handle)
{
	file_lock();
	
	fclose (sys_handles[handle]);
	sys_handles[handle] = 0;
	
	file_unlock();
}

void Sys_FileSeek (int handle, int position)
{
	file_lock();
	fseek (sys_handles[handle], position, SEEK_SET);
	file_unlock();
}

void Sys_FileSeekMore (int handle, int position, int whence)
{
	file_lock();
	fseek (sys_handles[handle], position, whence);
	file_unlock();
}

int Sys_FileRead (int handle, void *dest, int count)
{
	file_lock();
	int length_read = fread (dest, 1, count, sys_handles[handle]);
	if (length_read != count)
		Sys_Printf("read was too short\n");
		
	file_unlock();
	return length_read;
}

//int mmap_file_handle = -1;
//unsigned mmap_file_length;
//
//void* mmap_openfile(char *filename)
//{
//	mmap_file_length = Sys_FileOpenRead(filename, &mmap_file_handle);
//	
//	if (mmap_file_length < 0)
//	{
//		Sys_Printf("failed to load in mmapped file, %s\n", filename);
//		return NULL;
//	}
//	else
//	{
//		Sys_Printf("file %s successfully mmapped\n", filename);
//		return (void *)0x10000000;
//	}
//}
//
//void mmap_closefile(void)
//{
//	 mmap_file_length = 0;
//	 Sys_FileClose(mmap_file_handle);
//}
//
unsigned int mmap_handle(unsigned ea, int direction, int size)
{
}
//{
//	unsigned int data;
////	ea = ea - 0x10000000;
//	
//	if (size == 0)
//		size = 4;
//	
//	Sys_Printf("handle %08x, %d, %d\n", ea, direction, size);
//	
//	if (direction == 1)			//load
//	{
////		Sys_FileSeek(mmap_file_handle, ea);
////		Sys_FileRead(mmap_file_handle, &data, size);
//		data = ea;
//		
////		Sys_Printf("data is %08x\n", data);
//	}
//	else if (direction == 0)	//store
//	{
//	}
//	else
//		Sys_Error("\nmemory direction is %d!\n", direction);
//	
////	while(1);
//	return data;
//}

int Sys_FileWrite (int handle, void *data, int count) __attribute__ ((no_instrument_function));
int Sys_FileWrite (int handle, void *data, int count)
{
	file_lock();
	int res = fwrite (data, 1, count, sys_handles[handle]);
	file_unlock();
	return res;
}

int Sys_FileTime (char *path)
{
	file_lock();
	FILE    *f;
	
	f = fopen(path, "r");
	if (f)
	{
		fclose(f);
		file_unlock();
		return 1;
	}
	
	file_unlock();
	
	return -1;
}

unsigned short *frame_buffer_base;
unsigned short *console_base;
int frame_buffer_width;
int frame_buffer_height;

int blanks = 0;

void init_arm7(void)
{
	printf("Init ARM7...");
	quake_ipc_9to7_buf = (unsigned char *)malloc(100);
	quake_ipc_9to7_buf = (unsigned char *)((unsigned int)quake_ipc_9to7_buf | 0x400000);
	quake_ipc_7to9_buf = (unsigned char *)malloc(100);
	quake_ipc_7to9_buf = (unsigned char *)((unsigned int)quake_ipc_7to9_buf | 0x400000);
	
//	printf("9: 9->7 %08x 7->9 %08x\n", quake_ipc_9to7_buf, quake_ipc_7to9_buf);
	
	//send the init message
	quake_ipc_9to7->message_type = kInit;
	quake_ipc_9to7->message = 0xffffffff;
	
	//the ARM7 will now change the message to say it's ready
	while (quake_ipc_9to7->message == 0xffffffff);
	
//	printf("9: 7 has begun init\n");
	
	quake_ipc_9to7->message = (unsigned int)quake_ipc_9to7_buf;
	
	//it'll change this when it's read it
	while (quake_ipc_9to7->message != 0xffffffff);
	
//	printf("9: 7 has read 9to7\n");
	
	quake_ipc_9to7->message = (unsigned int)quake_ipc_7to9_buf;
	
	//it'll change this when it's read it
	while (quake_ipc_9to7->message == (unsigned int)quake_ipc_7to9_buf);
	
//	printf("9: 7 has read 7to9\n");
	
	quake_ipc_9to7->message = 0;
}

void stop_mp3(void)
{
#ifdef USE_MP3
	printf("stopping track\n");
	while(quake_ipc_9to7->message == 0xffffffff);
	
	quake_ipc_9to7->message_type = kStopMP3;
	quake_ipc_9to7->message = 0xffffffff;
	
	while(quake_ipc_9to7->message == 0xffffffff);
	
	while (decoder_stopped == 0);
#endif
}

void play_mp3(unsigned char track)
{
#ifdef USE_MP3
	if (decoder_stopped == 0)
		stop_mp3();
	
	decoder_stopped = 0;
	
	while(quake_ipc_9to7->message == 0xffffffff);
	printf("playing track %d\n", track);
	
	quake_ipc_9to7->message_type = kPlayMP3;
	sprintf((char *)quake_ipc_9to7_buf, "/quake soundtrack/0%d-AudioTrack 0%d.mp3", track, track);
	quake_ipc_9to7->message = 0xffffffff;
	
	while(quake_ipc_9to7->message == 0xffffffff);
#else
	decoder_stopped = 1;
#endif
}

void SCR_CheckDrawCenterString (void);
void SCR_DrawConsole (void);
void Draw_Crosshair(void);

int update_count = 0;

unsigned short *new_palette = NULL;
void ds_loadpalette(unsigned short *palette);

void vblank_handler(void)
{
#ifdef USE_QUAKE_CONSOLE
	update_count++;
	if (update_count > 10)
	{
		extern unsigned short *console_base;
		int count;
		for (count = 0; count < 128 * 128; count++)
			console_base[count] = 0;
		
		extern int extra;
		if (extra == 0)
			extra = 1;
		else
			extra = 0;
			
		SCR_CheckDrawCenterString ();
		SCR_DrawConsole ();
		Draw_Crosshair();
		
		update_count = 0;
	}
#endif

	if (new_palette != NULL)
	{
		ds_loadpalette(new_palette);
		new_palette = NULL;
	}
	
	handle_ipc();
	
	blanks++;
}

int hblanks = 0;

void hblank_handler(void)
{
//	unsigned short vcount = *((volatile unsigned short *)0x4000006);
	hblanks++;
//	if (vcount > 261)
//		hblanks = 0;
}

const double hblank_ms = 1.0 / 15.7343;
double Sys_FloatTime (void)
{
	static double t;
	
	t += 0.1;
	
	return t;
//	return hblanks * hblank_ms;
}

void ds_dcacheflush(void)
{
	DC_FlushAll();
}

float z_tri = 0.0f;
float z_quad = 0.0f;

int flush_val = 2;

float left_move = 0;
float up_move = 0;
float fov = 50;
float x_scale = 160;
float y_scale = 160;
float z_scale = 80;
float z_move = 0;
int palette_address = 0;

void ds_ortho(void)
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-100, 100, -70, 70, 1.0f, 0x7fff);
	glMatrixMode(GL_MODELVIEW);
}

void ds_perspective(void)
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(fov, 256.0 / 192.0, 4.0f, 0x7fff);
	glMatrixMode(GL_MODELVIEW);
}

void ds_begin_render(void)
{
#ifdef USE_3D
	glReset();
	glMatrixMode(GL_PROJECTION);
	
	ds_perspective();
	
	glMaterialf(GL_AMBIENT, RGB15(0,0,0));
	glMaterialf(GL_DIFFUSE, RGB15(0,0,0));
	glMaterialf(GL_SPECULAR, BIT(15) | RGB15(0,0,0));
	glMaterialf(GL_EMISSION, RGB15(0,0,0));

	//ds uses a table for shinyness..this generates a half-ass one
	glMaterialShinyness();
	
	glPolyFmt(POLY_ALPHA(31) | POLY_CULL_NONE);
	glMatrixMode(GL_TEXTURE);
	glIdentity();
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	
	glTranslatef(left_move, up_move, z_move);
	glScalef(x_scale, y_scale, z_scale);
	
//	glEnable(GL_ANTIALIAS);
#endif
}

void ds_end_render(void)
{
#ifdef USE_3D
	glPopMatrix(1);
	//glFlush();
	GFX_FLUSH = flush_val;
#endif
}

void ds_pushmatrix(void)
{
#ifdef USE_3D
	glPushMatrix();
#endif
}

void ds_popmatrix(void)
{
#ifdef USE_3D
	glPopMatrix(1);
#endif
}

void ds_translatef(float x, float y, float z)
{
#ifdef USE_3D
	glTranslatef(x, y, z);
#endif
}

void ds_scalef(float x, float y, float z)
{
#ifdef USE_3D
	glScalef(x, y, z);
#endif
}

void ds_rotateX(float angle)
{
#ifdef USE_3D
	glRotateX(angle);
#endif
}

void ds_rotateY(float angle)
{
#ifdef USE_3D
	glRotateY(angle);
#endif
}

void ds_rotateZ(float angle)
{
#ifdef USE_3D
	glRotateZ(angle);
#endif
}

void ds_usematrix3x3(float *up, float *right, float *forward)
{
#ifdef USE_3D
	int upf32[3], rightf32[3], forwardf32[3];
	int count;
	
	for (count = 0; count < 3; count++)
	{
		upf32[count] = floattof32(up[count]);
		rightf32[count] = floattof32(right[count]);
		forwardf32[count] = floattof32(forward[count]);
	}
	
//	MATRIX_MULT3x3 = upf32[0];
//	MATRIX_MULT3x3 = upf32[1];
//	MATRIX_MULT3x3 = upf32[2];
//	
//	MATRIX_MULT3x3 = rightf32[0];
//	MATRIX_MULT3x3 = rightf32[1];
//	MATRIX_MULT3x3 = rightf32[2];
//	
//	MATRIX_MULT3x3 = forwardf32[0];
//	MATRIX_MULT3x3 = forwardf32[1];
//	MATRIX_MULT3x3 = forwardf32[2];
	
	MATRIX_MULT3x3 = upf32[0];
	MATRIX_MULT3x3 = rightf32[0];
	MATRIX_MULT3x3 = forwardf32[0];
	
	MATRIX_MULT3x3 = upf32[1];
	MATRIX_MULT3x3 = rightf32[1];
	MATRIX_MULT3x3 = forwardf32[1];
	
	MATRIX_MULT3x3 = upf32[2];
	MATRIX_MULT3x3 = rightf32[2];
	MATRIX_MULT3x3 = forwardf32[2];
#endif
}

void ds_begin_triangle(void)
{
#ifdef USE_3D
	glBegin(GL_TRIANGLE);
#endif
}

void ds_begin_quad(void)
{
#ifdef USE_3D
	glBegin(GL_QUAD);
#endif
}

void ds_begin_triangle_strip(void)
{
#ifdef USE_3D
	glBegin(GL_TRIANGLE_STRIP);
#endif
}

void ds_end(void)
{
#ifdef USE_3D
	glEnd();
#endif
}

void ds_vertex3f(float x, float y, float z)
{
#ifdef USE_3D
	glVertex3f(x, y, z);
#endif
}

void ds_vertex3v16(short int x, short int y, short int z)
{
#ifdef USE_3D
	glVertex3v16(x, y, z);
#endif
}

void ds_color3f(float red, float green, float blue)
{
#ifdef USE_3D
	glColor3f(red, green, blue);
#endif
}

void ds_color3b(unsigned char red, unsigned char green, unsigned char blue)
{
#ifdef USE_3D
	glColor3b(red, green, blue);
#endif
}

void ds_texcoord2t16(short u, short v)
{
#ifdef USE_3D
	glTexCoord2t16(u, v);
#endif
}

void ds_texcoord2f(float u, float v)
{
#ifdef USE_3D
	glTexCoord2f(u, v);
#endif
}

void BindNoTexture(void);
void ResetMenuCache(void);

int vram_used = 0;
void ds_resettextures(void)
{
#ifdef USE_3D
	glResetTextures();
	vram_used = 0;
	
	BindNoTexture();
	ResetMenuCache();
	
	ds_begin_render();
	ds_end_render();
	ds_begin_render();
	ds_end_render();
#endif
}

int ds_gettexwidth(void)
{
#ifdef USE_3D
	int i;
	glGetInt(GL_GET_TEXTURE_WIDTH, &i);
	
	return i;
#else
	return 0;
#endif
}

int ds_gettexheight(void)
{
#ifdef USE_3D
	int i;
	glGetInt(GL_GET_TEXTURE_HEIGHT, &i);
	
	return i;
#else
	return 0;
#endif
}

extern unsigned int *textures;
extern unsigned int activeTexture;
void ds_bindtexture(int id)
{
#ifdef USE_3D
	glBindTexture(0, id);
#endif
}

int ds_gentexture(void)
{
#ifdef USE_3D
	int temp_texture;
	glGenTextures(1, &temp_texture);
	return temp_texture;
#endif
}

extern "C" int getNextPaletteSlot(unsigned short count, unsigned char format);

void ds_schedule_loadpalette(unsigned short *palette)
{
	new_palette = palette;
}

void ds_loadpalette(unsigned short *palette)
{
#ifdef USE_3D

	if (palette_address == 0)
	{
//		int addr = gluTexLoadPal(palette, 256, GL_RGB256);
		int addr = getNextPaletteSlot(256, GL_RGB256);
		  
		if( addr>=0 )
		{
			vramSetBankF(VRAM_F_LCD);
	 		swiCopy( palette, &VRAM_F[addr>>1] , 256 / 2 | COPY_MODE_WORD);
		 	vramSetBankF(VRAM_F_TEX_PALETTE);
		}
		
		if (addr == -1)
		{
			printf("couldn\'t load palette!");
			*(int *)0 = 0;
		}
		palette_address = addr;
	}
	else
//		glTexLoadPal(palette, 256, palette_address);
	{
		vramSetBankF(VRAM_F_LCD);
 		swiCopy( palette, &VRAM_F[palette_address>>1] , 256 / 2 | COPY_MODE_WORD);
	 	vramSetBankF(VRAM_F_TEX_PALETTE);
	}
		
	glColorTable(GL_RGB256, palette_address);
#endif
}

int tex_size(int size)
{
	switch (size)
	{
		case 8:
			size = TEXTURE_SIZE_8;
			break;
		case 16:
			size = TEXTURE_SIZE_16;
			break;
		case 32:
			size = TEXTURE_SIZE_32;
			break;
		case 64:
			size = TEXTURE_SIZE_64;
			break;
		case 128:
			size = TEXTURE_SIZE_128;
			break;
		case 256:
			size = TEXTURE_SIZE_256;
			break;
		case 512:
			size = TEXTURE_SIZE_512;
			break;
		case 1024:
			size = TEXTURE_SIZE_1024;
			break;
		default:
//			printf("invalid texture size, %d\n", size);
			size = -1;
			break;
	}
	return size;
}

int next_size_up(int size)
{
	if (size > 512)
		return 1024;
	if (size > 256)
		return 512;
	if (size > 128)
		return 256;
	if (size > 64)
		return 128;
	if (size > 32)
		return 64;
	if (size > 16)
		return 32;
	if (size > 8)
		return 16;
	if (size > 0)
		return 8;
	return 0;
}

unsigned int snapshot_activeTexture;
unsigned int *snapshot_nextBlock;
unsigned int snapshot_nextPBlock;
int snapshot_nameCount;
int snapshot_vram_used;

void ds_texture_snapshot(void)
{
#ifdef USE_3D
	extern unsigned int activeTexture;
	extern unsigned int *nextBlock;
	extern unsigned int nextPBlock;
	extern int nameCount;
	
	snapshot_activeTexture = activeTexture;
	snapshot_nextBlock = nextBlock;
	snapshot_nextPBlock = nextPBlock;
	snapshot_nameCount = nameCount;
	
	snapshot_vram_used = vram_used;
#endif
}

void ds_restore_snapshot(void)
{
#ifdef USE_3D
	extern unsigned int activeTexture;
	extern unsigned int *nextBlock;
	extern unsigned int nextPBlock;
	extern int nameCount;
	
	activeTexture = snapshot_activeTexture;
	nextBlock = snapshot_nextBlock;
	nextPBlock = snapshot_nextPBlock;
	nameCount = snapshot_nameCount;
	
	vram_used = snapshot_vram_used;
#endif
}

#define SSC_Disabled (0)
#define SSC_SDRAM (1)
#define SSC_CF (2)

#define SC_REG_UNLOCK	*(vu16*)(0x09FFFFFE)
static void SetSC_UNLOCK(int SSC)
{
  switch(SSC){
    case SSC_Disabled:
      SC_REG_UNLOCK = 0xA55A;
      SC_REG_UNLOCK = 0xA55A;
      SC_REG_UNLOCK = 0x0001;
      SC_REG_UNLOCK = 0x0001;
      break;
    case SSC_SDRAM:
      SC_REG_UNLOCK = 0xA55A;
      SC_REG_UNLOCK = 0xA55A;
      SC_REG_UNLOCK = 0x0005;
      SC_REG_UNLOCK = 0x0005;
      break;
    case SSC_CF:
      SC_REG_UNLOCK = 0xA55A;
      SC_REG_UNLOCK = 0xA55A;
      SC_REG_UNLOCK = 0x0003;
      SC_REG_UNLOCK = 0x0003;
  }
}
#undef SC_REG_UNLOCK

//void speed_test(void)
//{
//	SetSC_UNLOCK(SSC_SDRAM);
////	unsigned int *ptr_i = (unsigned int *)0x08000000;
////	unsigned short *ptr_s = (unsigned short *)0x08000000;
//
//	unsigned int *ptr_i = (unsigned int *)malloc(8096);
//	unsigned short *ptr_s = (unsigned short *)ptr_i;
//	
//	register unsigned int wr_i = 0x12;
//	volatile unsigned int start_time = *((u16 volatile *) 0x04000006);
//	
//	for (register int count = 0; count < (8192 >> 2); count += 8)
//	{
//		ptr_i[count] = wr_i;
//		ptr_i[count + 1] = wr_i;
//		ptr_i[count + 2] = wr_i;
//		ptr_i[count + 3] = wr_i;
//		ptr_i[count + 4] = wr_i;
//		ptr_i[count + 5] = wr_i;
//		ptr_i[count + 6] = wr_i;
//		ptr_i[count + 7] = wr_i;
//	}
//	
//	volatile unsigned int end_time = *((u16 volatile *) 0x04000006);
//	printf("%.2f megs/second (32-bit)\n", ((float)(60 * 262) / (end_time - start_time)) * (float)8 / 1024);
//	
//	register unsigned short wr_s = 0x12;
//	start_time = *((u16 volatile *) 0x04000006);
//
//	for (register int count = 0; count < (8192 >> 2); count += 8)
//	{
//		ptr_s[count] = wr_s;
//		ptr_s[count + 1] = wr_s;
//		ptr_s[count + 2] = wr_s;
//		ptr_s[count + 3] = wr_s;
//		ptr_s[count + 4] = wr_s;
//		ptr_s[count + 5] = wr_s;
//		ptr_s[count + 6] = wr_s;
//		ptr_s[count + 7] = wr_s;
//	}
//	
//	end_time = *((u16 volatile *) 0x04000006);
//	printf("%.2f megs/second (16-bit)\n", ((float)(60 * 262) / (end_time - start_time)) * (float)8 / 1024);
//	
//	while(1);
//
////	volatile unsigned int *ptr = (volatile unsigned int *)0x08000000;
////	unsigned int special = 0x12345678;
////	unsigned int read_back = 0;
////	
////	int count = 0;
////	while(1)
////	{
////		*ptr = special;
////		read_back = *ptr;
////		if (read_back != special)
////			break;
////		else
////			count++;
////		ptr++;
////	}
////	printf("%d words of goodness\n", count);
////	while(1);
//	SetSC_UNLOCK(SSC_CF);
//}

//from libnds, modified to do resizing on the fly
extern "C" uint32* getNextTextureSlot(int size);
extern "C" void glTexParameter(uint8 sizeX, uint8 sizeY, uint32* addr, int mode, uint32 param);
int glTexImage2DQuake(int target, int empty1, int type, int sizeX, int sizeY, int empty2, int param, uint8* texture,
	unsigned int rows, unsigned int cols, unsigned int realY, unsigned int realX) {
//---------------------------------------------------------------------------------
  uint32 size = 0;
  uint32* addr;
  uint32 vramTemp;

  size = 1 << (sizeX + sizeY + 6);
  

  switch (type) {
    case GL_RGB:
    case GL_RGBA:
      size = size << 1;
      break;
    case GL_RGB4:
      size = size >> 2;
      break;
    case GL_RGB16:
      size = size >> 1;
      break;

    default:
      break;
  }
  
  addr = getNextTextureSlot(size);
  
  if(!addr)
    return 0;

  // unlock texture memory
  vramTemp = vramSetMainBanks(VRAM_A_LCD,VRAM_B_LCD,VRAM_C_LCD,VRAM_D_LCD);

  if (type == GL_RGB) {
    // We do GL_RGB as GL_RGBA, but we set each alpha bit to 1 during the copy
    u16 * src = (u16*)texture;
    u16 * dest = (u16*)addr;
    
    glTexParameter(sizeX, sizeY, addr, GL_RGBA, param);
    
    while (size--) {
      *dest++ = *src | (1 << 15);
      src++;
    }
  } else {
    // For everything else, we do a straight copy
    glTexParameter(sizeX, sizeY, addr, type, param);

	if ((rows == 0) || (cols == 0))
		swiCopy((uint32*)texture, addr , size / 4 | COPY_MODE_WORD);
	else
	{
		memset(addr, 0, size);
		
		realX = realX >> 2;
		
		unsigned char *dest = (unsigned char *)addr;
		unsigned char *src = (unsigned char *)texture;
		
		for (int y = 0; y < rows; y++)
			memcpy(&addr[y * realX], &texture[y * cols], cols);
	}
  }

  vramRestoreMainBanks(vramTemp);

  return 1;
}

int ds_teximage2d(int sizeX, int sizeY, unsigned char* texture)
{
#ifdef USE_3D
	int tex_xsize, tex_ysize;
	int resize_xsize, resize_ysize;
	int do_resize = 0;
	
	if ((sizeX == 0) || (sizeY == 0))
	{
		printf("zero size texture!\n");
		return -1;
	}
	if ((sizeX > 1024) || (sizeY > 1024))
	{
		printf("texture is too big! %dx%d\n", sizeX, sizeY);
		return -1;
	}
	
	tex_xsize = tex_size(sizeX);
	tex_ysize = tex_size(sizeY);
	resize_xsize = sizeX;
	resize_ysize = sizeY;
	
	if ((tex_xsize == -1) || (tex_ysize == -1))
	{
		//we're gonna have to enlarge it as it's not a normal size
		if (tex_xsize == -1)
			resize_xsize = next_size_up(sizeX);
		if (tex_ysize == -1)
			resize_ysize = next_size_up(sizeY);

		if (resize_xsize * resize_ysize > 256 * 256)
			return -1;
		else
		{
//			int y;
//			memset((void *)resize, 0, resize_xsize * resize_ysize);
//			
//			for (y = 0; y < sizeY; y++)
//				memcpy(&resize[y * resize_xsize], &texture[y * sizeX], sizeX);
			
			do_resize = 1;
		}
	}
	
	vram_used += resize_xsize * resize_ysize;
		
#ifdef USE_DEBUGGER
	char text_buf[100];
	sprintf(text_buf, "%.2fk VRAM used (%.2fk), %.2fk total\n",
		(float)(sizeX * sizeY) / 1024, (float)(resize_xsize * resize_ysize) / 1024, (float)vram_used / 1024);
//	DEBUG_MSG(text_buf);
	printf(text_buf);
#endif
	tex_xsize = tex_size(resize_xsize);
	tex_ysize = tex_size(resize_ysize);
	
	if (!do_resize)
		return glTexImage2DQuake(0, 0, GL_RGB256, tex_xsize, tex_ysize, 0, TEXGEN_OFF | GL_TEXTURE_WRAP_S | GL_TEXTURE_WRAP_T, texture, 0, 0, 0, 0) - 1;
	else
	{
//		printf("using resized version, %dx%d->%dx%d\n", sizeX, sizeY, resize_xsize, resize_ysize);
		return glTexImage2DQuake(0, 0, GL_RGB256, tex_xsize, tex_ysize, 0, TEXGEN_OFF | GL_TEXTURE_WRAP_S | GL_TEXTURE_WRAP_T, texture, sizeY, sizeX, resize_ysize, resize_xsize) - 1;
	}
#endif
	return 0;
}

extern "C" void ds_soundsetup(unsigned char vol, unsigned char pan)
{
	setGenericSound(11025, vol, pan, 1);
}

extern "C" void ds_playsound(void *data, int length)
{
	playGenericSound(data, length);
}

void ds_stopallsounds(void)
{
	quake_ipc_9to7->message_type = kStopAllSounds;
	quake_ipc_9to7->message = 0xffffffff;
}

int real_main(int argc, unsigned char **argv)
{
	powerON(POWER_ALL);
  
#ifndef USE_3D
	// Use the main screen for output
 	videoSetMode(MODE_FB0 | DISPLAY_BG0_ACTIVE);
 	vramSetBankA(VRAM_A_LCD);
#else
    
//    videoSetMode(MODE_0_2D | DISPLAY_BG0_ACTIVE);
// 	vramSetBankA(VRAM_A_MAIN_BG);
 	
 	BG0_CR = BG_MAP_BASE(31);
 	videoSetMode(MODE_0_3D);
 	
 	vramSetBankA(VRAM_A_TEXTURE);
 	vramSetBankB(VRAM_B_TEXTURE);
 	vramSetBankC(VRAM_C_TEXTURE);
 	vramSetBankD(VRAM_D_TEXTURE);
//	vramSetBankC(VRAM_C_ARM7);
//	vramSetBankD(VRAM_D_ARM7);
#endif
	
#ifdef USE_QUAKE_CONSOLE
	videoSetModeSub(MODE_3_2D | DISPLAY_BG3_ACTIVE);
	vramSetBankH(VRAM_H_SUB_BG);
	SUB_BG3_CR = BG_BMP16_128x128 | BG_BMP_BASE(0);
	
	SUB_BG3_XDX = 1<<7;
 	SUB_BG3_YDY = 1<<7;
 	SUB_BG3_XDY = 0;
 	SUB_BG3_YDX = 0;
 	SUB_BG3_CX = 0;
 	SUB_BG3_CY = 0;
 #else
 	consoleDemoInit();
 #endif

	glViewPort(0,0,255,191);
	glClearColor(0,0,0);

	glClearDepth(0x7FFF);
	
//	lcdSwap();

	printf("Starting up\n");
	
#ifndef USE_DEBUGGER
	irqInit();
#endif
	printf("fat init...");
	fatInitDefault();
	printf("ok\n");

#ifdef USE_DEBUGGER
	printf("starting debugger\n");
	//start up the wireless, connect to the AP
	set_verbosity(VERBOSE_ERROR | VERBOSE_INFO/* | VERBOSE_TRACE | VERBOSE_MMAP*/);
	wireless_init(1);
	wireless_connect();

	printf("connecting...");
	//connect to the PC stub, initialise the debugger
	debugger_connect_tcp(192, 168, 1, 10);
	debugger_init();					//from here on in you can get the debugger's attention
	printf("ok\n");
	
	//this will check for new commands from the debugger - call this repeatedly
	//the first time it is call it will break until you decide to resume
	//you can call this more than once in your program
	user_debugger_update();			//the earlier you call it, the earlier you can pause your program
#endif
	
	irqSet(IRQ_VBLANK, vblank_handler);
	irqEnable(IRQ_VBLANK);
	irqSet(IRQ_HBLANK, hblank_handler);
	irqEnable(IRQ_HBLANK);
	
	int proper_argc = 3;
	char *proper_argv[5];
	char argv0[20], argv1[20], argv2[20], argv3[20], argv4[20];
	
	strcpy(argv0, "/_boot_mp.nds");
	strcpy(argv1, "-basedir");
	strcpy(argv2, "/");
	strcpy(argv3, "-game");
	strcpy(argv4, "qdqwav");
	
	proper_argv[0] = argv0;
	proper_argv[1] = argv1;
	proper_argv[2] = argv2;
	proper_argv[3] = argv3;
	proper_argv[4] = argv4;
	
	frame_buffer_base = VRAM_A;
	frame_buffer_width = SCREEN_WIDTH;
	frame_buffer_height = SCREEN_HEIGHT;
	console_base = BG_GFX_SUB;
	
	setGenericSound(11025, 127, 64, 1);
	
	init_arm7();
	
//	volatile int count = 0;
//	for (count = 0; count < 1000000; count++);
//	play_mp3(3);
	
//	while(1);
	
//	cygprofile_enable();
	
//	speed_test();
	quake_main(proper_argc, proper_argv);
		
	return 0;
}

int count_largest_block(void)
{
	int largest = 0;
	void *ptr = NULL;
	
	do
	{
		if (ptr != NULL)
		{
			free(ptr);
			ptr = NULL;
			largest++;
		}
			
		ptr = malloc(largest + 1);
		
	} while(ptr != NULL);
	
	return largest;
}

int count_largest_block_kb(void)
{
	int largest = 0;
	void *ptr = NULL;
	
	do
	{
		if (ptr != NULL)
		{
			free(ptr);
			ptr = NULL;
			largest += 1024;
		}
			
		ptr = malloc(largest + 1024);
		
	} while(ptr != NULL);
	
	return largest;
}

void print_top_four_blocks_kb(void)
{
	void *l1, *l2, *l3, *l4;
	
	printf("big free blocks, kb:\n");
	int size = count_largest_block_kb();
	l1 = malloc(size);
	printf("1: %.2f\n", (float)size / 1024);
	
	size = count_largest_block_kb();
	l2 = malloc(size);
	printf("2: %.2f\n", (float)size / 1024);
	
	size = count_largest_block_kb();
	l3 = malloc(size);
	printf("3: %.2f\n", (float)size / 1024);
	
	size = count_largest_block_kb();
	l4 = malloc(size);
	printf("4: %.2f\n", (float)size / 1024);
	
	free(l1);
	free(l2);
	free(l3);
	free(l4);
	
	printf("done\n");
}

void print_top_four_blocks(void)
{
	void *l1, *l2, *l3, *l4;
	
	printf("big free blocks, kb:\n");
	int size = count_largest_block();
	l1 = malloc(size);
	printf("1: %.2f\n", (float)size / 1024);
	
	size = count_largest_block();
	l2 = malloc(size);
	printf("2: %.2f\n", (float)size / 1024);
	
	size = count_largest_block();
	l3 = malloc(size);
	printf("3: %.2f\n", (float)size / 1024);
	
	size = count_largest_block();
	l4 = malloc(size);
	printf("4: %.2f\n", (float)size / 1024);
	
	free(l1);
	free(l2);
	free(l3);
	free(l4);
	
	printf("done\n");
}

void get_pen_pos(short *px, short *py)
{
	touchPosition touchXY = touchReadXY();
	*px = touchXY.px;
	*py = touchXY.py;
}

int main(int argc, unsigned char **argv)
{
//	asm volatile ( 
//	  "mov sp, #0x02400000\n" 
//	  "sub sp, sp, #4" 
//	);
	
	real_main(argc, argv);
	
	return 0;
}
